@@ -127,10 +127,10 @@ module Zip
     # Returns modification time by default.
     def time(component: :mtime)
       time =
-        if @extra['UniversalTime']
-          @extra['UniversalTime'].send(component)
-        elsif @extra['NTFS']
-          @extra['NTFS'].send(component)
+        if @extra[:universaltime]
+          @extra[:universaltime].send(component)
+        elsif @extra[:ntfs]
+          @extra[:ntfs].send(component)
         end
 
       # Standard time field in central directory has local time
@@ -155,13 +155,13 @@ module Zip
     # Sets modification time by default.
     def time=(value, component: :mtime)
       @dirty = true
-      unless @extra.member?('UniversalTime') || @extra.member?('NTFS')
-        @extra.create('UniversalTime')
+      unless @extra.member?(:universaltime) || @extra.member?(:ntfs)
+        @extra.create(:universaltime)
       end
 
       value = DOSTime.from_time(value)
       comp = "#{component}=" unless component.to_s.end_with?('=')
-      (@extra['UniversalTime'] || @extra['NTFS']).send(comp, value)
+      (@extra[:universaltime] || @extra[:ntfs]).send(comp, value)
       @time = value if component == :mtime
     end
 
@@ -179,7 +179,7 @@ module Zip
 
     # Does this entry return time fields with accurate timezone information?
     def absolute_time?
-      @extra.member?('UniversalTime') || @extra.member?('NTFS')
+      @extra.member?(:universaltime) || @extra.member?(:ntfs)
     end
 
     # Return the compression method for this entry.
@@ -200,7 +200,12 @@ module Zip
 
     # Does this entry use the ZIP64 extensions?
     def zip64?
-      !@extra['Zip64'].nil?
+      !@extra[:zip64].nil?
+    end
+
+    # Is this entry encrypted with AES encryption?
+    def aes?
+      !@extra[:aes].nil?
     end
 
     def file_type_is?(type) # :nodoc:
@@ -382,11 +387,12 @@ module Zip
 
       read_extra_field(extra, local: true)
       parse_zip64_extra(true)
+      parse_aes_extra
       @local_header_size = calculate_local_header_size
     end
 
     def pack_local_entry # :nodoc:
-      zip64 = @extra['Zip64']
+      zip64 = @extra[:zip64]
       [::Zip::LOCAL_ENTRY_SIGNATURE,
        @version_needed_to_extract, # version needed to extract
        @gp_flags, # @gp_flags
@@ -400,9 +406,17 @@ module Zip
        @extra ? @extra.local_size : 0].pack('VvvvvvVVVvv')
     end
 
-    def write_local_entry(io, rewrite: false) # :nodoc:
+    def write_local_entry(io, suppress_extra_fields: false, rewrite: false) # :nodoc:
       prep_local_zip64_extra
-      verify_local_header_size! if rewrite
+
+      # If we are rewriting the local header, then we verify that we haven't changed
+      # its size. At this point we have to keep extra fields if they are present.
+      if rewrite
+        verify_local_header_size!
+      elsif suppress_extra_fields
+        @extra.suppress_fields!(suppress_extra_fields)
+      end
+
       @local_header_offset = io.tell
 
       io << pack_local_entry
@@ -430,10 +444,7 @@ module Zip
         _, # diskNumberStart
         @internal_file_attributes,
         @external_file_attributes,
-        @local_header_offset,
-        @name,
-        @extra,
-        @comment = buf.unpack('VCCvvvvvVVVvvvvvVV')
+        @local_header_offset = buf.unpack('VCCvvvvvVVVvvvvvVV')
     end
 
     def set_ftype_from_c_dir_entry # :nodoc:
@@ -511,6 +522,7 @@ module Zip
       check_c_dir_entry_comment_size
       set_ftype_from_c_dir_entry
       parse_zip64_extra(false)
+      parse_aes_extra
     end
 
     def file_stat(path) # :nodoc:
@@ -559,7 +571,7 @@ module Zip
     end
 
     def pack_c_dir_entry # :nodoc:
-      zip64 = @extra['Zip64']
+      zip64 = @extra[:zip64]
       [
         @header_signature,
         @version, # version of encoding software
@@ -578,14 +590,11 @@ module Zip
         zip64 && zip64.disk_start_number ? 0xFFFF : 0, # disk number start
         @internal_file_attributes, # file type (binary=0, text=1)
         @external_file_attributes, # native filesystem attributes
-        zip64 && zip64.relative_header_offset ? 0xFFFFFFFF : @local_header_offset,
-        @name,
-        @extra,
-        @comment
+        zip64 && zip64.relative_header_offset ? 0xFFFFFFFF : @local_header_offset
       ].pack('VCCvvvvvVVVvvvvvVV')
     end
 
-    def write_c_dir_entry(io) # :nodoc:
+    def write_c_dir_entry(io, suppress_extra_fields: false) # :nodoc:
       prep_cdir_zip64_extra
 
       case @fstype
@@ -607,6 +616,7 @@ module Zip
         end
       end
 
+      @extra.suppress_fields!(suppress_extra_fields) if suppress_extra_fields
       io << pack_c_dir_entry
 
       io << @name
@@ -667,7 +677,7 @@ module Zip
                when 'file'
                  if name_is_directory?
                    raise ArgumentError,
-                         "entry name '#{newEntry}' indicates directory entry, but " \
+                         "entry name '#{@name}' indicates a directory entry, but " \
                          "'#{src_path}' is not a directory"
                  end
                  :file
@@ -677,7 +687,7 @@ module Zip
                when 'link'
                  if name_is_directory?
                    raise ArgumentError,
-                         "entry name '#{newEntry}' indicates directory entry, but " \
+                         "entry name '#{@name}' indicates a directory entry, but " \
                          "'#{src_path}' is not a directory"
                  end
                  :symlink
@@ -792,14 +802,28 @@ module Zip
       return unless zip64?
 
       if for_local_header
-        @size, @compressed_size = @extra['Zip64'].parse(@size, @compressed_size)
+        @size, @compressed_size = @extra[:zip64].parse(@size, @compressed_size)
       else
-        @size, @compressed_size, @local_header_offset = @extra['Zip64'].parse(
+        @size, @compressed_size, @local_header_offset = @extra[:zip64].parse(
           @size, @compressed_size, @local_header_offset
         )
       end
     end
 
+    def parse_aes_extra # :nodoc:
+      return unless aes?
+
+      if @extra[:aes].vendor_id != 'AE'
+        raise Error, "Unsupported encryption method #{@extra[:aes].vendor_id}"
+      end
+
+      unless ::Zip::AESEncryption::VERSIONS.include? @extra[:aes].vendor_version
+        raise Error, "Unsupported encryption style #{@extra[:aes].vendor_version}"
+      end
+
+      @compression_method = @extra[:aes].compression_method if ftype != :directory
+    end
+
     # For DEFLATED compression *only*: set the general purpose flags 1 and 2 to
     # indicate compression level. This seems to be mainly cosmetic but they are
     # generally set by other tools - including in docx files. It is these flags
@@ -830,7 +854,7 @@ module Zip
       # If we already have a ZIP64 extra (placeholder) then we must fill it in.
       if zip64? || @size.nil? || @size >= 0xFFFFFFFF || @compressed_size >= 0xFFFFFFFF
         @version_needed_to_extract = VERSION_NEEDED_TO_EXTRACT_ZIP64
-        zip64 = @extra['Zip64'] || @extra.create('Zip64')
+        zip64 = @extra[:zip64] || @extra.create(:zip64)
 
         # Local header always includes size and compressed size.
         zip64.original_size = @size || 0
@@ -844,7 +868,7 @@ module Zip
       if (@size && @size >= 0xFFFFFFFF) || @compressed_size >= 0xFFFFFFFF ||
          @local_header_offset >= 0xFFFFFFFF
         @version_needed_to_extract = VERSION_NEEDED_TO_EXTRACT_ZIP64
-        zip64 = @extra['Zip64'] || @extra.create('Zip64')
+        zip64 = @extra[:zip64] || @extra.create(:zip64)
 
         # Central directory entry entries include whichever fields are necessary.
         zip64.original_size = @size if @size && @size >= 0xFFFFFFFF
